#ifdef WIN32

#include <Windows.h>
#include <FL/Fl.H>
#include "FL/fl_ask.H"
#include "FL/x.H"
#include "FL/fl_draw.H"
#include "src/Fl_Font.H"
#include <stdarg.h>
#include <string.h>
#include <direct.h>
#include <stdlib.h>
#include <stdio.h>

/* included functions */
void MG_fl_draw(const char *txt, int x, int y);
void *MG_PrepareCopy(int w, int h);
void MG_CompleteCopy(void *oldflgc);
extern "C" {
	void my_bell(void);
	int mysystem(const char *command);
	char *get_prog_dir(void);
}
char **getutf8args(int *pargc);
void draw_rotated_string(const char *strDraw, int x, int y, float degrees /*clockwise*/);

/* extern vars & functions */
extern HDC fl_gc;
#if FL_MAJOR_VERSION*100 + FL_MINOR_VERSION >= 103
extern Fl_Font_Descriptor* fl_fontsize;
#else
extern Fl_FontSize* fl_fontsize;
#endif
extern char *wchartoutf8(const wchar_t *in);

void MG_fl_draw(const char *txt, int x, int y)
/* it is reprogrammed because in FLTK 1.3 text is output one char at a time
 and this is not convenient when doing copy/paste
 */
{
	if(txt == NULL) return;
	int n = strlen(txt);
	COLORREF oldColor = SetTextColor(fl_gc, fl_RGB());
	if (fl_fontsize) SelectObject(fl_gc, fl_fontsize->fid);
	WCHAR *wout = new WCHAR[n+5];
	int newl = MultiByteToWideChar(CP_UTF8, 0, txt, n, wout, n + 5);
	TextOutW(fl_gc, x, y, wout, newl);
	delete wout;
	SetTextColor(fl_gc, oldColor);
}


void *MG_PrepareCopy(int w, int h)
/* returns NULL iff error
 */
{
	HDC oldflgc = fl_gc;
	int factor = 32; // empirically set
	RECT rect; rect.left = 0; rect.top = 0; rect.right=w*factor; rect.bottom=h*factor;
	fl_gc = CreateEnhMetaFile (NULL, NULL, &rect, NULL);
	if (fl_gc != NULL) {
		SetTextAlign(fl_gc, TA_BASELINE|TA_LEFT);
		SetBkMode(fl_gc, TRANSPARENT);
		return oldflgc ? oldflgc : fl_gc;
    } 
	else {
		fl_gc = oldflgc;
		return NULL;
	}
}


void MG_CompleteCopy(void *oldflgc)
{
	if(oldflgc == fl_gc) oldflgc = NULL;
	HENHMETAFILE hmf = CloseEnhMetaFile (fl_gc);
	if ( hmf != NULL ) {
		if ( OpenClipboard (NULL) ){
			EmptyClipboard ();
			SetClipboardData (CF_ENHMETAFILE, hmf);
			CloseClipboard ();
		}
		DeleteEnhMetaFile(hmf);
	}
	DeleteDC(fl_gc);
	fl_gc = (HDC)oldflgc;
}


static int   abortPrint;
static PRINTDLG      pd;
static HDC           hPr = NULL;
static int           prerr;
extern HWND fl_window;

static void WIN_SetupPrinterDeviceContext(HDC prHDC)
{
	if ( !prHDC ) return;
	
	fl_window = 0;
	fl_gc = prHDC;
	SetMapMode(prHDC, MM_ANISOTROPIC);
	SetTextAlign(prHDC, TA_BASELINE|TA_LEFT);
	SetBkMode(prHDC, TRANSPARENT);	
	HWND screenHwnd = GetDesktopWindow();
	HDC  screenDC   = GetDC( screenHwnd );
	
	/* NOTE:  for video devices, the logical inch IS NOT == physical inch! */
	SetWindowExtEx(prHDC, GetDeviceCaps(screenDC, LOGPIXELSX),
				   GetDeviceCaps(screenDC, LOGPIXELSY), NULL);
	SetViewportExtEx(prHDC, GetDeviceCaps(prHDC, LOGPIXELSX),
					 GetDeviceCaps(prHDC, LOGPIXELSY), NULL);
	
	ReleaseDC(screenHwnd, screenDC);
}


int MG_StartPrinting (int pagecount, int *frompage, int *topage)
// returns 0 iff OK
{
	DWORD       commdlgerr;
	DOCINFO     di;
	char        docName [256];
	int err = 0;
	
	abortPrint = FALSE;
	memset (&pd, 0, sizeof (PRINTDLG));
	pd.lStructSize = sizeof (PRINTDLG);
	pd.hwndOwner = GetForegroundWindow();
	pd.Flags = PD_RETURNDC | PD_USEDEVMODECOPIESANDCOLLATE | PD_NOSELECTION;
	pd.nMinPage = 1;
	pd.nMaxPage = pagecount;
	if (PrintDlg (&pd) != 0) {
		hPr = pd.hDC;
		if (hPr != NULL) {
			WIN_SetupPrinterDeviceContext (hPr);
			strcpy (docName, "Seaview");
			memset(&di, 0, sizeof(DOCINFO));
			di.cbSize = sizeof (DOCINFO);
			di.lpszDocName = (LPCSTR) docName;
			prerr = StartDoc (hPr, &di);
			if (prerr < 1) {
				abortPrint = TRUE;
				fl_alert ("StartDoc error %d", prerr);
				err = 1;
			}
		} else {
			commdlgerr = CommDlgExtendedError ();
			fl_alert ("Unable to create print context, error %lu",
						 (unsigned long) commdlgerr);
			err = 1;
		}
	} else {
		err = 1;
	}
	if(!err) {
		if((pd.Flags & PD_PAGENUMS) != 0 ) {
			*frompage = pd.nFromPage;
			*topage = pd.nToPage;
			}
		else {
			*frompage = 1;
			*topage = pagecount;
			}
		}
	return err;
}

void MG_EndPrinting (void)
{
	if (hPr != NULL) {
		if (! abortPrint) {
			prerr = EndDoc (hPr);
			if (prerr < 0) {
				fl_alert ("EndDoc error %d", prerr);
			}
		}
		DeleteDC (hPr);
		if (pd.hDevMode != NULL) {
			GlobalFree (pd.hDevMode);
		}
		if (pd.hDevNames != NULL) {
			GlobalFree (pd.hDevNames);
		}
	}
}


int MG_ClipPrintingRect(int x, int y, int w, int h)
{
	long left, right, top, bottom;
	POINT    pt;
		
	left = x - 1; right = x + w;
	top = y - 1; bottom = y + h ;
	
	if (hPr == NULL) return FALSE;
	
	pt.x = left;
	pt.y = bottom;
	LPtoDP(hPr, &pt, 1);
	left   = pt.x;
	bottom = pt.y;
	
	pt.x = right;
	pt.y = top;
	LPtoDP(hPr, &pt, 1);
	right = pt.x;
	top   = pt.y;
	
    HRGN hRgnClip = CreateRectRgn (left, top, right, bottom);
    SelectClipRgn (hPr, hRgnClip);
    DeleteObject (hRgnClip);
	return TRUE;
}


int MG_PrintingRect(int *x, int *y, int *w, int *h)
{
	POINT         physPageSize;
	POINT         pixelsPerInch;
		
	*x = *y = 0;
	
	if (hPr == NULL) return FALSE;
	
	physPageSize.x = GetDeviceCaps(hPr, HORZRES);
	physPageSize.y = GetDeviceCaps(hPr, VERTRES);
	DPtoLP(hPr, &physPageSize, 1);
	*w = physPageSize.x - *x + 1;
	*h = physPageSize.y - *y + 1;
	pixelsPerInch.x = GetDeviceCaps(hPr, LOGPIXELSX);
	pixelsPerInch.y = GetDeviceCaps(hPr, LOGPIXELSY);
	DPtoLP(hPr, &pixelsPerInch, 1);
	*x += (pixelsPerInch.x / 4);
	*w -= (pixelsPerInch.x / 2);
	*y += (pixelsPerInch.y / 4);
	*h -= (pixelsPerInch.y / 2);

	return TRUE;
}

int MG_StartPage (void)
{
	int  rsult;
	
	rsult = TRUE;
	if (hPr != NULL) {
		WIN_SetupPrinterDeviceContext (hPr);
		prerr = StartPage (hPr);
		if (prerr < 0) {
			fl_alert ("StartPage error %d", prerr);
			rsult = FALSE;
		}
		fl_window = 0;
		fl_gc = hPr;
	}
	return rsult;
}

int MG_EndPage (void)
{
	int  rsult;
	
	rsult = TRUE;
	if (hPr != NULL) {
		prerr = EndPage (hPr);
		if (prerr < 0) {
			abortPrint = TRUE;
			fl_alert ("EndPage error %d", prerr);
			rsult = FALSE;
		}
	} else {
		rsult = FALSE;
	}
	return rsult;
}


int mysystem(const char *command)
/*
 command contains "cmd.exe /c prog.exe args"  or  "prog.exe args"
 the first form is compulsory if args contain "> outfile"
 prog is searched in dir where calling program was launched, current dir,
 windows system dir, windows dir, dirs in PATH.
 returns 0 iff OK
 */
{
	static char buf[100];
	STARTUPINFO info;
	PROCESS_INFORMATION pi;
	int retval;
	
	GetStartupInfo(&info);
	retval = CreateProcess(NULL, (char *)command,0,0,0,0,0,_getcwd(buf,sizeof(buf)),&info,&pi);
	if(retval == 0) return 1;
	do	{
		Fl::check();
		retval = WaitForSingleObject(pi.hProcess, 500);
	}
	while(retval == WAIT_TIMEOUT);
	return 0;
}

char *get_prog_dir(void)
/* returns path of dir where calling program was launched */
{
	static char *dir = NULL;
	char *line, *p;
	int l;
	
	if(dir == NULL) {
		line = wchartoutf8(GetCommandLineW());
		//fprintf(OUT,"retour GetCommandLine line=%s\n",line);fflush(OUT);
		/* extract 1st word of line or first group delimited by " " */
		if(*line == '"') { line++; p = strchr(line, '"'); }
		else p = strchr(line, ' '); 
		if(p == NULL) p = line + strlen(line);
		l = p - line;
		while( l > 0 && line[l - 1] != '\\' ) l--;
		if(l > 0) {
			dir = (char *)malloc(l + 1);
			memcpy(dir, line, l); dir[l] = 0;
		}
	}
	return dir;
}


void my_bell(void)
{
	Beep(0,0);
} 


char **getutf8args(int *pargc)
{
	int argc;
	wchar_t **wargv = CommandLineToArgvW(GetCommandLineW(), &argc);
	char **argv = (char **)malloc(argc * sizeof(char *));
	for (int count = 0; count < argc; count++) {
		argv[count] = strdup(wchartoutf8(wargv[count]));
	}
	*pargc = argc;
	return argv;
}


void draw_rotated_string(const char *strDraw, int x, int y, float degrees /*clockwise*/)
{
	int n = strlen(strDraw);
	WCHAR *wout = new WCHAR[n+5];
	int newl = MultiByteToWideChar(CP_UTF8, 0, strDraw, n, wout, n + 5);
	LOGFONT lf;
	memset(&lf, 0, sizeof(LOGFONT));
	int rank = fl_font() / 4;
	SelectObject(fl_gc, fl_fontsize->fid);//necessary to put selected fltk's font in fl_gc
	GetTextFace(fl_gc, sizeof(lf.lfFaceName), lf.lfFaceName);

	if(fl_font() < 12) {//only the first 12 fonts have bold/italic
		rank = fl_font() % 4;
		lf.lfWeight = (rank == 1 || rank == 3) ? FW_BOLD : FW_NORMAL;
		lf.lfItalic = (rank == 2 || rank == 3) ? 1 : 0;
	}
	lf.lfHeight = - fl_size();

	// Specify the angle to draw line
	lf.lfEscapement = - (LONG)(degrees*10);
	lf.lfOrientation = lf.lfEscapement;
	// Select the new font created
	HFONT hFont = (HFONT) CreateFontIndirect(&lf);
	HFONT hPrevFont = (HFONT) SelectObject(fl_gc, hFont);

	// Draw text to GC with color
	SetTextColor(fl_gc, fl_RGB());
	TextOutW(fl_gc, x, y, wout, newl);
	SelectObject(fl_gc, hPrevFont);
	DeleteObject(hFont);
	delete wout;
}


/*
 void add_to_path(char *dir)
 /* ajoute a la variable PATH la directory dir
 teste si deja dedans ou si vide *
 {
 char *buffer;
 int l;
 
 if(dir == NULL || strlen(dir) == 0) return;
 buffer = (char *)malloc(10000);
 l = GetEnvironmentVariable("PATH", buffer, 10000);
 if(l == 0) *buffer = 0;
 if(strstr(buffer, dir) != NULL) return;
 if(*buffer != 0) strcat(buffer, ";");
 strcat(buffer, dir);
 SetEnvironmentVariable("PATH", buffer);
 free(buffer);
 }
 
 
 char **split_args(char *all, int *pargc)
 {
 char *p, *q, **argv;
 int l, argc = 0;
 
 // compter les arguments
 p = all;
 while(*p != 0) {
 argc++;
 if(*p == '"') p = strchr(p+1, '"');
 else p = strchr(p, ' ');
 if(p == NULL) break;
 p++;
 while(*p == ' ') p++;
 }
 argv = (char **)malloc(argc * sizeof(char *));
 *pargc = argc;
 // separer les arguments, enlever les "" encadrantes si presentes
 p = all; argc = 0;
 while(*p != 0) {
 if(*p == '"') { p++; q = strchr(p, '"'); }
 else q = strchr(p, ' ');
 if(q == NULL) q = p + strlen(p);
 l = q - p; 
 argv[argc] = (char *)malloc(l + 1);
 memcpy(argv[argc], p, l); argv[argc][l] = 0;
 if(*q == 0) break;
 p = q + 1;
 while(*p == ' ') p++;
 argc++;
 }
 return argv;
 }
 */

#endif // WIN32
